%{
/*
 *  ippl.y - functions parsing the configuration file
 *
 *  Copyright (C) 1998-2000 Etienne Bernard
 *  *Small* modifications for UDP by Hugo Haas, 1999
 *  Now resolves IP addresses at parsing time whenever possible,
 *    Etienne Bernard 1999
 *  Keeps hostname for refreshing. Added "expire" rule. Hugo Haas 1999
 *  Added a "detailed" and a "noresolve" rule. Hugo Haas 1999
 *  Added a "log-in" rule. Hugo Haas 1999
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <netdb.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>

#include "defines.h"
#include "configuration.h"
#include "filter.h"
#include "log.h"

void yyerror(char *);
void logError(char *, int);
int yylex(void);

const char *proto;

extern int goBackground;

extern struct filterEntry *filter;

extern int lineNumber;

/* User used to run threads */
char *safeUser = NULL;

/* Expiration of DNS data */
unsigned int dnsExpire;

/* Protocols logged */
unsigned short logProtocols;

/* Should name resolving be done? */
unsigned short resolveProtocols;

/* Logging format for each protocol */
unsigned short icmpFormat;
unsigned short tcpFormat;
unsigned short udpFormat;

/* Should we log end of TCP connections ? */
unsigned short logClosing;

/* Should we use ident requests? */
unsigned short useIdent;

/* Logging information */
extern logDevice *icmpLog, *tcpLog, *udpLog;

struct logInfo switches;
%}

%union {
  long longval;
  char * stringval;
  struct portRange_struct rangeval;
  struct portRanges_struct portdesc;
  union logInfo_union loginfoval;
  struct filterEntry * entryval;
  struct {
    long protocol;
    union logInfo_union loginfoval;
  } protocolval;
  struct {
    struct hostDesc_struct fromval;
    struct hostDesc_struct toval;
  } fromtoval;
  struct hostDesc_struct hostdesc;
}

%token<stringval> IP HOSTMASK IDENTIFIER FILENAME
%token<longval> NUMBER

%token LOGFORMAT DETAILED SHORT NORMAL RESOLVE NORESOLVE IDENT NOIDENT
%token LOGCLOSING NOLOGCLOSING RUN RUNAS EXPIRE LOG_IN LOG IGNORE FROM
%token TO TYPE PORT SRCPORT OPTION COMMA ICMP TCP UDP ALL RANGE SLASH
%token EOL

%type<entryval> Rule
%type<longval> Action ProtoList ProtoItem LogFormat
%type<rangeval> Range
%type<hostdesc> From To
%type<fromtoval> FromTo
%type<protocolval> Protocol
%type<rangeval> Port SrcPort
%type<loginfoval> Type
%type<stringval> Filename
%type<portdesc> PortDesc

%start Input
%%

Input:
          /* Empty */
        | Input Line
        ;

Line:
          EOL

        | IDENT EOL
          { useIdent = 1; }
        | NOIDENT EOL
          { useIdent = 0; }

        | RESOLVE ProtoList EOL
          { resolveProtocols |= $2; }
        | NORESOLVE ProtoList EOL
          { resolveProtocols &= ~$2; }

        | LOGCLOSING EOL
          { logClosing = 1; }
        | NOLOGCLOSING EOL
          { logClosing = 0; }

        | LOGFORMAT LogFormat ProtoList EOL
          {
            if ($3 & GET_ICMP)
              icmpFormat = $2;
            if ($3 & GET_TCP)
              tcpFormat = $2;
            if ($3 & GET_UDP)
              udpFormat = $2;
          }

        | RUN ProtoList EOL
          { logProtocols |= $2; }

        | RUNAS IDENTIFIER EOL
          {
            struct passwd *argument = getpwnam($2);
            if (argument != NULL) {
              free(safeUser);
              safeUser = $2;
            }
            else {
              logError("unknown user.", lineNumber);
              free($2);
            }
          }

        | EXPIRE NUMBER EOL
          { dnsExpire = $2; }

        | LOG_IN Login EOL

        | Rule EOL
           { if ($1) { $1->next = filter; filter = $1; } }
        | error EOL
          {
            logError("ignoring invalid line.", lineNumber-1);
          }
        ;

ProtoList:
          ProtoItem
        | ProtoList ProtoItem
          { $$ = $1 | $2; }
        ;

ProtoItem:
          ICMP
          { $$ = GET_ICMP; }
        | TCP
          { $$ = GET_TCP;  }
        | UDP
          { $$ = GET_UDP;  }
        | ALL
          { $$ = (GET_ICMP | GET_TCP | GET_UDP); }
        | error
          {
            logError("unknown protocol.", lineNumber); yyclearin;
            $$ = 0;
          }
        ;

LogFormat:
          NORMAL
          { $$ = LOGFORMAT_NORMAL; }
        | SHORT
          { $$ = LOGFORMAT_SHORT; }
        | DETAILED
          { $$ = LOGFORMAT_DETAILED; }
        ;

Login:
          ICMP Filename
            { if ($2 && goBackground) {
                if (icmpLog) icmpLog->close(icmpLog);
                free(icmpLog);
                icmpLog = newFileDevice($2);
                free($2);
              }
            }
        | TCP Filename
            { if ($2 && goBackground) {
                if (tcpLog) tcpLog->close(tcpLog);
                free(tcpLog);
                tcpLog = newFileDevice($2);
                free($2);
              }
            }
        | UDP Filename
            { if ($2 && goBackground) {
                if (udpLog) udpLog->close(udpLog);
                free(udpLog);
                udpLog = newFileDevice($2);
                free($2);
              }
            }
        | ALL Filename
          {
            if ($2 && goBackground) {
              if (icmpLog) icmpLog->close(icmpLog);
              if (tcpLog) tcpLog->close(tcpLog);
              if (udpLog) udpLog->close(udpLog);
              free(icmpLog); free(tcpLog); free(udpLog);
              icmpLog = newFileDevice($2);
              tcpLog = newFileDevice($2);
              udpLog = newFileDevice($2);
              free($2);
            }
          }
        | error
          { logError("unknown protocol.", lineNumber); yyclearin; }
        ;

Filename:
          FILENAME
          { $$ = $1; }
        | error
          {
            logError("incorrect filename.", lineNumber); yyclearin;
            $$ = NULL;
          }
        ;

Rule:
          Action
          {
            switches.log = -1;
            switches.ident = useIdent;
            switches.resolve = -1;
            switches.logFormat = -1;
            switches.logClosing = logClosing;
          } 
          Switches Protocol FromTo
          {
            $$ = (struct filterEntry *)malloc(sizeof(struct filterEntry));
            $$->log = $1;
            $$->ident = switches.ident;
            $$->logClosing = switches.logClosing;
            if (switches.resolve == -1) {
              $$->resolve = resolveProtocols & $4.protocol;
            } else {
              $$->resolve = switches.resolve;
            }
            if (switches.logFormat == -1) {
              switch ($4.protocol) {
              case IPPROTO_ICMP:
                $$->logFormat = icmpFormat; break;
              case IPPROTO_TCP:
                $$->logFormat = tcpFormat; break;
              case IPPROTO_UDP:
                $$->logFormat = udpFormat; break;
              }
            } else {
              $$->logFormat = switches.logFormat;
            }
            $$->protocol = $4.protocol;
            $$->logInfo = $4.loginfoval;
            $$->toDesc = $5.toval;
            $$->fromDesc = $5.fromval;
          }
        ;

Action:
          LOG     { $$ = 1; }
        | IGNORE  { $$ = 0; }
        ;

Switches:
        
        | OPTION SwitchList
        ;

SwitchList:
          Switch
        | SwitchList COMMA Switch
        ;

Switch:
          IDENT      { switches.ident = 1; }
        | NOIDENT    { switches.ident = 0; }
        | RESOLVE    { switches.resolve = 1; }
        | NORESOLVE  { switches.resolve = 0; }
        | SHORT      { switches.logFormat = LOGFORMAT_SHORT; }
        | NORMAL     { switches.logFormat = LOGFORMAT_NORMAL; }
        | DETAILED   { switches.logFormat = LOGFORMAT_DETAILED; }
        | LOGCLOSING { switches.logClosing = 1; }
        | NOLOGCLOSING { switches.logClosing = 0; }
        ;

Protocol:
          ICMP Type
          { $$.protocol = IPPROTO_ICMP; $$.loginfoval = $2;}
        | TCP
          { proto = "tcp"; }
          PortDesc
          { $$.protocol = IPPROTO_TCP; $$.loginfoval.portRanges = $3; }
        | UDP
          { proto = "udp"; }
          PortDesc
          { $$.protocol = IPPROTO_UDP; $$.loginfoval.portRanges = $3; }
        ;

Type:
          { $$.icmpType = NO_ICMP_TYPE; }
        | TYPE NUMBER { $$.icmpType = $2; }
        ;

PortDesc:
          { $$.dst.rangeMin = NOT_DEFINED;
            $$.dst.rangeMax = NOT_DEFINED;
            $$.src.rangeMin = NOT_DEFINED;
            $$.src.rangeMax = NOT_DEFINED; }
        | Port
          { $$.dst = $1;
            $$.src.rangeMin = NOT_DEFINED;
            $$.src.rangeMax = NOT_DEFINED; }
        | SrcPort
          { $$.dst.rangeMin = NOT_DEFINED;
            $$.dst.rangeMax = NOT_DEFINED;
            $$.src = $1; }
        | Port SrcPort
          { $$.dst = $1;
            $$.src = $2; }
        | SrcPort Port
          { $$.src = $1;
            $$.dst = $2; }
        ;

Port:
         PORT Range
          { $$ = $2; }
       ;

SrcPort:
          SRCPORT Range
          { $$ = $2; }
       ;

Range:
          NUMBER               { $$.rangeMin = $$.rangeMax = $1; }
	| IDENTIFIER           {
                                 struct servent *s = getservbyname($1, proto);
                                 if (s)
                                   $$.rangeMin = $$.rangeMax = ntohs(s->s_port);
                                 else {
                                   $$.rangeMin = $$.rangeMax = NOT_DEFINED;
                                   logError("unknown service.", lineNumber);
                                   YYERROR;
                                 }
                                 free($1);
                               }
        | RANGE NUMBER         { $$.rangeMin = NOT_DEFINED;
                                 $$.rangeMax = $2;                    }
        | NUMBER RANGE         { $$.rangeMin = $1;
                                 $$.rangeMax = NOT_DEFINED;           }
        | NUMBER RANGE NUMBER  { $$.rangeMin = $1; $$.rangeMax = $3; }
        ;

FromTo:
          {
            $$.fromval.address = $$.fromval.mask = 0;
            $$.fromval.hostName = $$.fromval.hostMask = NULL;
            $$.toval.address = $$.toval.mask = 0;
            $$.toval.hostName = $$.toval.hostMask = NULL;
          }
        | From
          {
            $$.fromval = $1;
            $$.toval.address = $$.toval.mask = 0;
            $$.toval.hostName = $$.toval.hostMask = NULL;
          }
        | From To
          {
            $$.fromval = $1;
            $$.toval = $2;
          }
        | To
          {
            $$.toval = $1;
            $$.fromval.address = $$.fromval.mask = 0;
            $$.fromval.hostName = $$.fromval.hostMask = NULL;
          }
        | To From
          {
            $$.toval = $1;
            $$.fromval = $2;
          }
        ;

From:
          FROM IP
            {
#ifdef HAVE_INET_ATON
              struct in_addr inp;
              inet_aton($2, &inp);
              $$.address = inp.s_addr;
#else
              $$.address = inet_addr($2);
#endif
              $$.mask = 0xffffffff;
              $$.hostName = NULL;
              $$.hostMask = NULL;
              free($2);
            }
        | FROM IP SLASH IP
            {
#ifdef HAVE_INET_ATON
              struct in_addr inp;
              inet_aton($2, &inp);
              $$.address = inp.s_addr;
              inet_aton($4, &inp);
              $$.mask = inp.s_addr;
#else
              $$.address = inet_addr($2);
              $$.mask = inet_addr($4);
#endif
              $$.hostName = NULL;              
              $$.hostMask = NULL;
              free($2); free($4);
            }
        | FROM IP SLASH NUMBER
            {
#ifdef HAVE_INET_ATON
              struct in_addr inp;
              inet_aton($2, &inp);
              $$.address = inp.s_addr;
#else
              $$.address = inet_addr($2);
#endif
              $$.mask = (1 << $4)-1;
              $$.hostName = NULL;              
              $$.hostMask = NULL;
              free($2);
            }
        | FROM HOSTMASK
            {
              if (strchr($2, '*')) {
                $$.hostMask = $2;
                $$.hostName = NULL;
                $$.address = 0;
                $$.mask = 0;
              } else {
                struct hostent * host;
                if ((host = gethostbyname((const char *)$2)) == 0) {
                  $$.address = 0;
                  $$.mask = 0;
                  logError("cannot resolve hostname.", lineNumber);
                } else {
                  memcpy(&($$.address), host->h_addr, host->h_length);
                  $$.mask = 0xffffffff;
                };
                $$.hostName = $2;
                $$.hostMask = NULL;
              }
            }
        | FROM IDENTIFIER
            {
              struct hostent * host;
              if ((host = gethostbyname((const char *)$2)) == 0) {
                $$.address = 0;
                $$.mask = 0;
                logError("cannot resolve hostname.", lineNumber);
              } else {
                memcpy(&($$.address), host->h_addr, host->h_length);
                $$.mask = 0xffffffff;
              };
              $$.hostName = $2;
              $$.hostMask = NULL;
            }
        ;

To:
          TO IP
            {
#ifdef HAVE_INET_ATON
              struct in_addr inp;
              inet_aton($2, &inp);
              $$.address = inp.s_addr;
#else
              $$.address = inet_addr($2);
#endif
              $$.mask = 0xffffffff;
              $$.hostName = NULL;
              $$.hostMask = NULL;
              free($2);
            }
        | TO HOSTMASK
            {
              if (strchr($2, '*')) {
                $$.address = $$.mask = 0;
                logError("wildcards are not allowed in a \"TO\" rule.", lineNumber);
              } else {
                struct hostent * host;
                if ((host = gethostbyname((const char *)$2)) == 0) {
                  $$.address = 0;
                  $$.mask = 0;
                  logError("cannot resolve hostname.", lineNumber);
                } else {
                  memcpy(&($$.address), host->h_addr, host->h_length);
                  $$.mask = 0xffffffff;
                };
                $$.hostName = $2;
                $$.hostMask = NULL;
              }
            }
        | TO IDENTIFIER
            {
              struct hostent * host;
              if ((host = gethostbyname((const char *)$2)) == 0) {
                $$.address = $$.mask = 0;
                logError("cannot resolve hostname.", lineNumber);
              } else {
                memcpy(&($$.address), host->h_addr, host->h_length);
                $$.mask = 0xffffffff;
              };
              $$.hostName = $2;
              $$.hostMask = NULL;
            }
        ;

%%

void yyerror(char *s) {
  return;
}

/* Print an error using the logging mechanism */
void logError(char *s, int l) {
  extern logDevice *mainLog;

  mainLog->doLog(mainLog, "CFG: Parse error line %d: %s", l, s);
}
