/*
 *  log.c - Logging mechanism
 *
 *  Copyright (C) 1999-2000 Hugo Haas & Etienne Bernard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#ifdef DEBUG
#include <assert.h>
#endif

#include "defines.h"
#include "log.h"

extern logDevice *mainLog;

/*****************************/
/* Syslog
 */

/* Log into syslog */
void logSyslog(logDevice *ld, char *format, ...) {
  char buffer[BUFFER_SIZE];
  va_list msg;

#ifdef DEBUG
  assert(ld);
#endif

  va_start(msg, format);
  vsnprintf(buffer, BUFFER_SIZE, format, msg);
  va_end(msg);

  syslog(LOGLEVEL, buffer);
}

/* Dummy close of a logging device */
void dummyClose(logDevice *ld) { }

/* Create a new syslog logging device */
logDevice *newSyslogDevice(int level) {
  logDevice *ld;

  ld = (logDevice *)malloc(sizeof(logDevice));
  ld->devInfo.syslogLevel = level;
  ld->doLog = logSyslog;
  ld->close = dummyClose;

  return ld;
}


/*****************************/
/* File
 */

/* Log a message */
void logMessage(int fd, time_t *timestamp, char *msg) {
  char date[27];

  snprintf(date, 27, asctime(localtime(timestamp)));
  write(fd, date+4, strlen(date)-10);
  write(fd, " ", 1);
  write(fd, msg, strlen(msg));
  write(fd, "\n", 1);
}

/* Variables used for last message repetition */
char lastMessage[BUFFER_SIZE] = "";
time_t lastRepeat;
unsigned int repeats = 0;

/* Log into a file */
void logFile(logDevice *ld, char *format, ...) {
  char buffer[BUFFER_SIZE], repeatsMessage[40];
  time_t curr = time(NULL);
  va_list msg;

#ifdef DEBUG
  assert(ld);
#endif

  va_start(msg, format);
  vsnprintf(buffer, BUFFER_SIZE, format, msg);
  va_end(msg);

  if (!strncmp(lastMessage, buffer, BUFFER_SIZE)) {
    repeats++;
    lastRepeat = curr;
    return;
  }

  if (repeats > 0) {
    snprintf(repeatsMessage, 40, "last message repeated %d time(s)", repeats);
    logMessage(ld->devInfo.fileDescriptor, &lastRepeat, repeatsMessage);
    repeats = 0;
  }

  strncpy(lastMessage, buffer, BUFFER_SIZE);

  logMessage(ld->devInfo.fileDescriptor, &curr, buffer);
}

/* Close a file logging device */
void closeFile(logDevice *ld) {
#ifdef DEBUG
  assert(ld);
#endif
  close(ld->devInfo.fileDescriptor);
}

/* Create a new file logging device */
logDevice *newFileDevice(char *filename) {
  int fd;
  logDevice *ld;

  fd = open(filename, O_WRONLY | O_APPEND | O_CREAT, 0640);
  if (!fd) {
    mainLog->doLog(mainLog, "Cannot open log file %s.", filename);
    return NULL;
  }
  ld = (logDevice *)malloc(sizeof(logDevice));
  ld->devInfo.fileDescriptor = fd;
  ld->doLog = logFile;
  ld->close = closeFile;

  return ld;
}

/*****************************/
/* StdOut
 */

/* Create a new stdout logging device */
logDevice *newStdOutDevice() {
  logDevice *ld;

  ld = (logDevice *)malloc(sizeof(logDevice));
  ld->devInfo.fileDescriptor = 1;
  ld->doLog = logFile;
  ld->close = dummyClose;

  return ld;
}
