/*
 *  logging.c - Functions to filter and log
 *
 *  Copyright (C) 1999-2000 Hugo Haas & Etienne Bernard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <netinet/in.h>

#include "filter.h"
#include "logging.h"
#include "thread.h"
#include "configuration.h"
#include "utils.h"
#include "ident.h"

#define ipOptions(ihl) ((ihl) == 5) ? "" : " (IP opts)"

/* Threads */
pthread_t logThread;

/* Mutexes and conditions */
pthread_mutex_t logMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t logCond = PTHREAD_COND_INITIALIZER;

/* Log list item */
struct logEntry {
  struct logInfo info;
  struct logEntry *next;
};
struct logEntry *logList = NULL;

/* Main logging loop */
void logLoop(struct passwd *user) {
  struct logEntry *entry;

#ifdef __OS_LINUX__
  /* Switch user */
  if (user) setuid(user->pw_uid);
#endif

  /* Enable cancelation */
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  /* Get mutex */
  pthread_mutex_lock(&logMutex);

  /* Infinite loop */
  while (1) {
    /* Wait for something to log */
    while (logList == NULL) {
      pthread_cond_wait(&logCond, &logMutex);
    }

    /* Log the packet */
    while (logList) {
      entry = logList;
      logList = logList->next;

      /* Really log the packet */
      logPacket(entry->info);

      free(entry);
    }
  }

  /* Release the mutex - never reached. */
  pthread_mutex_unlock(&logMutex);
}

/* Add packet into the logList */
void recordPacket(struct logInfo info) {
  struct logEntry *entry;

  /* Get the mutex */
  pthread_mutex_lock(&logMutex);

  /* Manipulate data structures */
  entry = (struct logEntry *)malloc(sizeof(struct logEntry));
  entry->info = info;
  entry->next = logList;
  logList = entry;

  /* Signal that there is something to log */
  pthread_cond_signal(&logCond);

  /* Release the mutex */
  pthread_mutex_unlock(&logMutex);
}

/* Start logging thread */
void startLogThread(struct passwd *user) {
  runThread(&logThread, (void *(*)(void *))logLoop, (void *)user);
}

/* Stop logging thread */
void stopLogThread(void) {
  pthread_cancel(logThread);
}

/* Log the packet */
void logPacket(struct logInfo info) {
  switch (info.protocol) {
  case IPPROTO_ICMP:
    logICMP(info); break;
  case IPPROTO_UDP:
    logUDP(info); break;
  case IPPROTO_TCP:
    logTCP(info); break;
  }
}

/* Log an ICMP packet */
void logICMP(struct logInfo info) {
  char details[DETAILS_LENGTH];
  char remoteHost[HOST_LENGTH];
  char icmpType[TYPE_LENGTH];
  char icmpCode[TYPE_LENGTH];

  *details = '\0';
  hostPrint(remoteHost, info.from, info.resolve);
  icmpTypeLookup(icmpType, icmpCode, info.srcPort, info.dstPort);

  switch (info.logFormat) {
  case LOGFORMAT_SHORT:
    info.logDevice->doLog(info.logDevice, "ICMP %s%s - %s", icmpType, icmpCode, remoteHost);
    break;
  case LOGFORMAT_NORMAL:
  case LOGFORMAT_DETAILED:
    getDetails(details, info.from, 0, info.to, 0);
    info.logDevice->doLog(info.logDevice, "ICMP message type %s%s from %s%s",
                          icmpType, icmpCode, remoteHost, ipOptions(info.ihl));
    break;
  }
}

/* Log an UDP packet */
void logUDP(struct logInfo info) {
  char details[DETAILS_LENGTH];
  char remoteHost[HOST_LENGTH];
  char service[SERVICE_LENGTH];

  *details = '\0';
  hostPrint(remoteHost, info.from, info.resolve);
  serviceLookup("udp", service, info.dstPort);

  switch (info.logFormat) {
  case LOGFORMAT_SHORT:
    info.logDevice->doLog(info.logDevice, "UDP %s - %s", service, remoteHost);
    break;
  case LOGFORMAT_NORMAL:
    info.logDevice->doLog(info.logDevice, "%s UDP datagram from %s", service, remoteHost);
    break;
  case LOGFORMAT_DETAILED:
    getDetails(details, info.from, info.srcPort, info.to, info.dstPort);
    info.logDevice->doLog(info.logDevice, "%s UDP datagram from %s%s%s",
                          service, remoteHost, details, ipOptions(info.ihl));
    break;
  }
}

/* Log a TCP packet */
void logTCP(struct logInfo info) {
  char details[DETAILS_LENGTH];
  char remoteHost[HOST_LENGTH];
  char service[SERVICE_LENGTH];
  char *username = NULL;

  *details = '\0';
  hostPrint(remoteHost, info.from, info.resolve);
  serviceLookup("tcp", service, info.dstPort);

  if (info.ident && info.dstPort != IDENT_PORT)
    username = getIdent(info.dstPort, info.srcPort, info.to, info.from);

  if (info.closing) {
    switch (info.logFormat) {
    case LOGFORMAT_SHORT:
      if (username) {
        info.logDevice->doLog(info.logDevice, "TCP %s closed - %s@%s", service, username, remoteHost);
        free(username);
      } else
        info.logDevice->doLog(info.logDevice, "TCP %s closed - %s", service, remoteHost);
      break;
    case LOGFORMAT_NORMAL:
      if (username) {
        info.logDevice->doLog(info.logDevice, "%s connection closed from %s@%s", service, username, remoteHost);
        free(username);
      } else
        info.logDevice->doLog(info.logDevice, "%s connection closed from %s", service, remoteHost);
      break;
    case LOGFORMAT_DETAILED:
      getDetails(details, info.from, info.srcPort, info.to, info.dstPort);
      if (username) {
        info.logDevice->doLog(info.logDevice, "%s connection closed from %s@%s%s%s",
                              service, username, remoteHost, details, ipOptions(info.ihl));
        free(username);
      } else
        info.logDevice->doLog(info.logDevice, "%s connection closed from %s%s%s",
                              service, remoteHost, details, ipOptions(info.ihl));
      break;
    }
  } else {
    switch (info.logFormat) {
    case LOGFORMAT_SHORT:
      if (username) { 
        info.logDevice->doLog(info.logDevice, "TCP %s - %s@%s", service, username, remoteHost);
        free(username);
      } else
        info.logDevice->doLog(info.logDevice, "TCP %s - %s", service, remoteHost);
      break;
    case LOGFORMAT_NORMAL:
      if (username) {
        info.logDevice->doLog(info.logDevice, "%s connection attempt from %s@%s", service, username, remoteHost);
        free(username);
      } else
        info.logDevice->doLog(info.logDevice, "%s connection attempt from %s", service, remoteHost);
      break;
    case LOGFORMAT_DETAILED:
      getDetails(details, info.from, info.srcPort, info.to, info.dstPort);
      if (username) {
        info.logDevice->doLog(info.logDevice, "%s connection attempt from %s@%s%s%s",
                              service, username, remoteHost, details, ipOptions(info.ihl));
        free(username);
      } else
        info.logDevice->doLog(info.logDevice, "%s connection attempt from %s%s%s",
                              service, remoteHost, details, ipOptions(info.ihl));
      break;
    }
  }
}
